class FlarumForum(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)


    # Main:
    @property
    def id(self) -> str:
        try:
            return self['data']['id']
        except KeyError:
            return None

    @property
    def type(self) -> str:
        try:
            return self['data']['type']
        except KeyError:
            return None


    # Attributes:
    @property
    def title(self) -> str:
        try:
            return self['data']['attributes']['title']
        except KeyError:
            return None

    @property
    def description(self) -> str:
        try:
            return self['data']['attributes']['description']
        except KeyError:
            return None

    @property
    def showLanguageSelector(self) -> bool:
        try:
            return self['data']['attributes']['showLanguageSelector']
        except KeyError:
            return None

    @property
    def baseUrl(self) -> str:
        try:
            return self['data']['attributes']['base_url']
        except KeyError:
            return None

    @property
    def basePath(self) -> str:
        try:
            return self['data']['attributes']['basePath']
        except KeyError:
            return None

    @property
    def debug(self) -> bool:
        try:
            return self['data']['attributes']['debug']
        except KeyError:
            return None

    @property
    def apiUrl(self) -> str:
        try:
            return self['data']['attributes']['apiUrl']
        except KeyError:
            return None

    @property
    def welcomeTitle(self) -> str:
        try:
            return self['data']['attributes']['welcomeTitle']
        except KeyError:
            return None

    @property
    def welcomeMessage(self) -> str:
        try:
            return self['data']['attributes']['welcomeMessage']
        except KeyError:
            return None

    @property
    def themePrimaryColor(self) -> str:
        try:
            return self['data']['attributes']['themePrimaryColor']
        except KeyError:
            return None

    @property
    def themeSecondaryColor(self) -> str:
        try:
            return self['data']['attributes']['themeSecondaryColor']
        except KeyError:
            return None

    @property
    def logoUrl(self) -> str:
        try:
            return self['data']['attributes']['logoUrl']
        except KeyError:
            return None

    @property
    def faviconUrl(self) -> str:
        try:
            return self['data']['attributes']['faviconUrl']
        except KeyError:
            return None

    @property
    def headerHtml(self) -> str:
        try:
            return self['data']['attributes']['headerHtml']
        except KeyError:
            return None

    @property
    def footerHtml(self) -> str:
        try:
            return self['data']['attributes']['footerHtml']
        except KeyError:
            return None

    @property
    def allowSignUp(self) -> bool:
        try:
            return self['data']['attributes']['allowSignUp']
        except KeyError:
            return None

    @property
    def defaultRoute(self) -> str:
        try:
            return self['data']['attributes']['defaultRoute']
        except KeyError:
            return None

    # Permissions:
    @property
    def canViewDiscussions(self) -> bool:
        try:
            return self['data']['attributes']['canViewDiscussions']
        except KeyError:
            return None

    @property
    def canStartDiscussions(self) -> bool:
        try:
            return self['data']['attributes']['canStartDiscussions']
        except KeyError:
            return None

    @property
    def canViewUserList(self) -> bool:
        try:
            return self['data']['attributes']['canViewUserList']
        except KeyError:
            return None

    @property
    def canViewFlags(self) -> bool:
        try:
            return self['data']['attributes']['canViewFlags']
        except KeyError:
            return None

    @property
    def guidelinesUrl(self) -> str:
        try:
            return self['data']['attributes']['guidelinesUrl']
        except KeyError:
            return None

    @property
    def minPrimaryTags(self) -> int:
        try:
            return self['data']['attributes']['minPrimaryTags']
        except KeyError:
            return None

    @property
    def maxPrimaryTags(self) -> int:
        try:
            return self['data']['attributes']['maxPrimaryTags']
        except KeyError:
            return None

    @property
    def minSecondaryTags(self) -> int:
        try:
            return self['data']['attributes']['minSecondaryTags']
        except KeyError:
            return None

    @property
    def maxSecondaryTags(self) -> int:
        try:
            return self['data']['attributes']['maxSecondaryTags']
        except KeyError:
            return None

    @property
    def canSaveDrafts(self) -> bool:
        try:
            return self['data']['attributes']['canSaveDrafts']
        except KeyError:
            return None

    @property
    def canScheduleDrafts(self) -> bool:
        try:
            return self['data']['attributes']['canScheduleDrafts']
        except KeyError:
            return None

    @property
    def canSelectBestAnswerOwnPost(self) -> bool:
        try:
            return self['data']['attributes']['canSelectBestAnswerOwnPost']
        except KeyError:
            return None

    @property # TODO: Move to extensions
    def maxUserBioLength(self) -> int:
        try:
            return self['data']['attributes']['fof-user-bio.maxLength']
        except KeyError:
            return None

    @property
    def canRequestUsername(self) -> bool:
        try:
            return self['data']['attributes']['canRequestUsername']
        except KeyError:
            return None

    @property
    def canRequestNickname(self) -> bool:
        try:
            return self['data']['attributes']['canRequestNickname']
        except KeyError:
            return None



from datetime import datetime
from typing import Generator

from pyflarum.date_conversions import flarum_to_datetime


class FlarumNotification(dict):
    """
        This represents a Flarum notification.
    """

    def __init__(self, *args, **kwargs):
        """
            Initializes the notification data.
        """

        if args[0]['type'] == 'notifications':
            dict.__init__(self, *args, **kwargs)

        else:
            raise TypeError("This is not a Flarum notification.")


    # Main:
    @property
    def id(self) -> str:
        """The notification's ID"""

        try:
            return self['id']
        except KeyError:
            return None

    @property
    def type(self) -> str:
        """The type of the notification. This should be always `notifications`, otherwise a `TypeError` is raised."""

        try:
            return self['type']
        except KeyError:
            return None

    @property
    def attributes(self) -> dict:
        """A raw data `dict` of notification's attributes."""

        try:
            return self['attributes']
        except KeyError:
            return None

    @property
    def contentType(self) -> str:
        """Notification's content type. This varies with different notification types."""

        try:
            return self.attributes['contentType']
        except KeyError:
            return None

    @property
    def content(self) -> dict:
        """The notification's content. Currently, this is always `None`, except for new post notification."""

        try:
            return self.attributes['content']
        except KeyError:
            return None
    
    @property
    def createdAt(self) -> datetime:
        """The `datetime` when the notification was created."""

        try:
            return flarum_to_datetime(self.attributes['createdAt'])
        except KeyError:
            return None

    @property
    def isRead(self) -> bool:
        try:
            return self.attributes['isRead']
        except KeyError:
            return None
    
    @property
    def relationships(self) -> dict:
        """The raw `dict` data of notification's relationships."""

        try:
            return self['relationships']
        except KeyError:
            return None

    @property
    def from_type(self) -> int:
        """
            The user/something's(?) type that triggered the notification. At the moment, this is always `"users"`
            (because only users can trigger notifications), but I kept this here in case an update or extension
            changes this behavior, so you can check who/what triggered the notification.
        """

        try:
            return self.relationships['fromUser']['data']['type']
        except KeyError:
            return None

    @property
    def from_id(self) -> int:
        """The user/something's(?) ID that triggered the notification."""

        try:
            return self.relationships['fromUser']['data']['id']
        except KeyError:
            return None

    @property
    def subject_type(self) -> int:
        """The subject's type this notification was triggered in. This can be `"discussions"` or `"posts"`."""

        try:
            return self.relationships['subject']['data']['type']
        except KeyError:
            return None

    @property
    def subject_id(self) -> int:
        """
            The subject's ID this notification was triggered in. A subject can be a discussion or post. Check `subject_type`.
            So, in case the notification was triggered by a post, then this will be that post's ID.
        """

        try:
            return self.relationships['subject']['data']['id']
        except KeyError:
            return None


class FlarumNewPostNotification(FlarumNotification):
    """
        Represents the "new post" Flarum notification (when a new post is made in a discussion the user follows).
    """

    def __init__(self, *args, **kwargs):
        """
            Initializes the Flarum new post notification's data. If `contentType` is not `newPost`, a `TypeError` is raised.
        """

        if args[0]['attributes']['contentType'] == "newPost":
            FlarumNotification.__init__(self, *args, **kwargs)

        else:
            raise TypeError("This is not a new post notification.")

    @property
    def postNumber(self) -> str:
        """The post number in a discussion this notification was triggered in. This is NOT post's ID."""

        try:
            return self.content['postNumber']
        except KeyError:
            return None


class FlarumPostMentionedNotification(FlarumNotification):
    """
        Represents the "post mentioned" Flarum notification (when someone mentions your post by replying to it).
    """

    def __init__(self, *args, **kwargs):
        """
            Initializes the Flarum post mentioned notification's data. If `contentType` is not `postMentioned`, a `TypeError` is raised.
        """

        if args[0]['attributes']['contentType'] == "postMentioned":
            FlarumNotification.__init__(self, *args, **kwargs)

        else:
            raise TypeError("This is not a post mentioned notification.")

    @property
    def replyNumber(self) -> str:
        """The post number in a discussion this notification was triggered in. This is NOT post's ID."""

        try:
            return self.content['replyNumber']
        except KeyError:
            return None


class FlarumNotifications(dict):
    """
        A collection of bulk obtained notifications.
    """

    def __init__(self, *args, **kwargs):
        """
            Initializes the notifications' data.
        """

        dict.__init__(self, *args, **kwargs)


    @property
    def rawData(self) -> dict:
        """
            Raw JSON data of all notifications, as it came from the API.
        """

        try:
            return self
        except KeyError:
            return None

    @property
    def new_posts(self) -> Generator[FlarumNewPostNotification, None, None]:
        """
            Represents all notifications about new posts in a discussion you're following.
        """

        try:
            for data in self['data']:
                if data['attributes']['contentType'] == "newPost":
                    yield FlarumNewPostNotification(data)

        except KeyError:
            return None
    
    @property
    def post_mentions(self) -> Generator[FlarumPostMentionedNotification, None, None]:
        """
            Represents all notifications about replies on your post/when
            someone mentions your post (by replying to it).
        """

        try:
            for data in self['data']:
                if data['attributes']['contentType'] == "postMentioned":
                    yield FlarumPostMentionedNotification(data)

        except KeyError:
            return None
    
    @property
    def user_mentions(self) -> Generator[FlarumNotification, None, None]:
        """
            Represents all notifications about your mentions (when someone mentions you).
        """

        try:
            for data in self['data']:
                if data['attributes']['contentType'] == "userMentioned":
                    yield FlarumNotification(data)

        except KeyError:
            return None
    
    @property
    def liked_posts(self) -> Generator[FlarumNotification, None, None]:
        """
            Represents all notifications about liked posts (when someone likes one of your posts).
        """

        try:
            for data in self['data']:
                if data['attributes']['contentType'] == "postLiked":
                    yield FlarumNotification(data)

        except KeyError:
            return None



from typing import Generator

from pyflarum.classes.flarum.FlarumUser import FlarumUser


class FlarumUsers(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)


    @property
    def firstLink(self) -> str:
        try:
            return self['links']['first']
        except KeyError:
            return None
    
    @property
    def nextLink(self) -> str:
        try:
            return self['links']['next']
        except KeyError:
            return None
    
    @property
    def all(self) -> Generator[FlarumUser, None, None]:
        try:
            for data in self['data']:
                user = FlarumUser(data)
                yield user
        except KeyError:
            return None



from typing import Union
from datetime import datetime


def flarum_to_datetime(flarum_dt: Union[str, None]=None) -> datetime:
    """
        Converts Flarum's datetime string to Python's datetime object.
        Doesn't convert if the parameter is already a datetime object.
        
        Flarum's datetime format is `%Y-%m-%dT%H:%M:%S%z`
    """

    if flarum_dt is None:
        return None

    if type(flarum_dt) == datetime:
        return flarum_dt

    return datetime.strptime(flarum_dt, '%Y-%m-%dT%H:%M:%S%z')


def datetime_to_flarum(dt: Union[datetime, None]=None) -> str:
    """
        Converts Python's datetime object to Flarum's datetime string.
        Doesn't convert if the parameter is already a string.
        
        Flarum's datetime format is `%Y-%m-%dT%H:%M:%S%z`
    """

    if dt is None:
        return None

    if type(dt) == str:
        return dt

    return dt.strftime('%Y-%m-%dT%H:%M:%S%z')



from datetime import datetime
from typing import Generator

from pyflarum.classes.extensions.FriendsOfFlarum_Polls import FoF_Poll
from pyflarum.classes.flarum.FlarumPosts import FlarumPost

from pyflarum.date_conversions import flarum_to_datetime
from pyflarum.classes.extensions.Flarum_Tags import FlarumPostDiscussionTagsData

from pyflarum.exceptions import FlarumError


class FlarumDiscussion(dict):
    def __init__(self,
                 title: str=None,
                 content: str=None,
                 tags: FlarumPostDiscussionTagsData=None,
                 poll: FoF_Poll=None,
                 raw: dict=None):
        """
            Initializes the discussion's data.
        """

        if raw is None:
            if title is None or content is None or type(title) == dict:
                raise FlarumError("Discussion title and content is required.")

            template = {
                "data": {
                    "type": "discussions",
                    "attributes": {
                        "title": title,
                        "content": content
                    },
                    "relationships": {}
                }
            }

            if poll is not None:
                if type(poll) == FoF_Poll:
                    template['attributes']['poll'] = poll
                else:
                    raise FlarumError('The poll must be a "FoF_Poll" object.')

            if tags is not None:
                if type(tags) == FlarumPostDiscussionTagsData:
                    template['relationships']['tags'] = tags
                else:
                    raise FlarumError("Tag IDs must be a list.")

            dict.__init__(self, template)

        else:
            if type(raw) == dict:
                dict.__init__(self, raw)
            
            else:
                raise FlarumError("This is not a Flarum discussion.")


    # Main:
    @property
    def id(self) -> str:
        try:
            return self['id']
        except KeyError:
            return None

    @property
    def type(self) -> str:
        try:
            return self['type']
        except KeyError:
            return None

    @property
    def slug(self) -> str:
        try:
            return self['attributes']['slug']
        except KeyError:
            return None

    # Attributes:
    @property
    def title(self) -> str:
        try:
            return self['attributes']['title']
        except KeyError:
            return None
    
    @property
    def commentCount(self) -> int:
        try:
            return self['attributes']['commentCount']
        except KeyError:
            return None
    
    @property
    def participantCount(self) -> int:
        try:
            return self['attributes']['participantCount']
        except KeyError:
            return None
    
    @property
    def createdAt(self) -> datetime:
        try:
            return flarum_to_datetime(self['attributes']['createdAt'])
        except KeyError:
            return None

    # Included:
    @property
    def posts(self) -> Generator[FlarumPost, None, None]:
        try:
            for post in self['included']:
                if post['type'] == 'posts':
                    yield FlarumPost(post)
                else:
                    continue

        except KeyError:
            return None

    @property
    def poll(self) -> FoF_Poll:
        poll_data = FoF_Poll()

        try:
            for included in self['included']:

                if included['type'] == "polls":
                    poll_data.update(included)
                    continue

                if included['type'] == "poll_options":
                    poll_data['_options'] = []
                    poll_data['_options'].append(included)
                    continue

        except KeyError:
            raise

        return poll_data


class FlarumDiscussions(dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)


    @property
    def firstLink(self) -> str:
        try:
            return self['links']['first']
        except KeyError:
            return None
    
    @property
    def nextLink(self) -> str:
        try:
            return self['links']['next']
        except KeyError:
            return None
    
    @property
    def all_discussions(self) -> Generator[FlarumDiscussion, None, None]:
        try:
            for data in self['data']:
                discussion = FlarumDiscussion(raw=data)
                yield discussion

        except KeyError:
            return None



from datetime import datetime
from typing import Generator, Union

from pyflarum.classes.flarum.FlarumNotifications import FlarumNotifications

from pyflarum.classes.FlarumSession import FlarumSession

from pyflarum.classes.flarum.FlarumDiscussions import FlarumDiscussion, FlarumDiscussions

from pyflarum.classes.flarum.FlarumPosts import FlarumPost

from pyflarum.classes.flarum.FlarumUser import FlarumUser
from pyflarum.classes.flarum.FlarumUsers import FlarumUsers

from pyflarum.classes.extensions.FriendsOfFlarum_Polls import FoF_Poll, FoF_PollOption

from pyflarum.exceptions import FlarumError, FlarumPermissionError, FlarumUnauthenticatedError


class FlarumMyUser(FlarumSession):
    """
        The main class for performing all Flarum related User actions that the `pyflarum` package offers.
    """

    def __init__(self, forum_url: str, username: Union[str, None]=None, password: Union[str, None]=None, cache: bool=True, cache_expire_after: int=300, proxies: dict=None):
        """
            The initializator for `FlarumMyUser()`.
        """

        FlarumSession.__init__(self, forum_url, username, password, cache, cache_expire_after, proxies)

        # API URLs:
        self.ENDPOINTS = {
            "api_discussions_url": f"{self.forum_url}/api/discussions",
            "api_posts_url": f"{self.forum_url}/api/posts",
            "self.api_users_url": f"{self.forum_url}/api/users"
        }

    @property
    def user(self) -> Union[FlarumUser, None]:
        if not self.isAuthenticated:
            raise FlarumUnauthenticatedError("Cannot obtain self user data when not authenticated.")

        user = None

        try:
            user = self.get_user_by_username(self.username)

        except:
            raise

        finally:
            return user
    

    def get_notifications(self) -> Union[FlarumNotifications, None]:
        notifications = None

        if not self.isAuthenticated:
            raise FlarumUnauthenticatedError("Cannot obtain notifications when not authenticated.")
        
        try:
            raw_notifications = self.session.get(f"{self.forum_url}/api/notifications").json() # type: dict

            if 'errors' in raw_notifications:
                raise FlarumError(raw_notifications)

            notifications = FlarumNotifications(raw_notifications)

            return notifications

        except KeyError:
            raise

        finally:
            return notifications



    def create_discussion(self, discussion: FlarumDiscussion) -> bool:
        """
            Creates a discussion.
        """

        if not self.isAuthenticated:
            raise FlarumUnauthenticatedError("Cannot create discussion when not authenticated.")
        
        if discussion is None or type(discussion) != FlarumDiscussion:
            raise FlarumError("discussion parameter must be a 'FlarumDiscussion' object.")

        discussion_response = None

        try:
            discussion_response = self.session.post(url=self.ENDPOINTS['api_discussions_url'], json=discussion).json() # type: dict

            if 'errors' in discussion_response:
                raise FlarumError(discussion_response)

            discussion_response = FlarumDiscussion(raw=discussion_response)

        except:
            raise
        
        finally:
            return discussion_response
    
    def mark_notifications_as_read(self) -> bool:
        self.session.post(f"{self.forum_url}/api/notifications/read")
    

    def like_post(self, post: FlarumPost) -> bool:
        """
            Likes a post.

            - `post_id`: The post ID to like.

            #### Example:
            ```python
            post = flarum_session.get_post_by_id(12)
            flarum
            ```
        """

        if post.canLike:
            self.session.patch(url=f"{self.forum_url}/api/posts/{post.id}", json={"data":{"type":"posts","id":"50","attributes":{"isLiked":True}}})

        else:
            FlarumPermissionError(f"You can't like post {post.id}")

        return True


    def unlike_post(self, post: FlarumPost) -> bool:
        """
            Unlikes a liked post.

            - `post_id`: The post ID to unlike.

            #### Example:
            ```python
            post = flarum_session.get_post_by_id(12)
            flarum_session.unlike_post(post)
            ```
        """

        self.session.patch(url=f"{self.forum_url}/api/posts/{post.id}", json={"data":{"type":"posts","id":"50","attributes":{"isLiked":False}}})

        return True

    def reply_to_discussion(self, discussion_id: int, content: str) -> Union[FlarumPost, None]:
        post = None

        template = {
            "data": {
                "type":"posts",
                "attributes": {
                    "content": content
                },
                "relationships": {
                    "discussion": {
                        "data": {
                            "type": "discussions",
                            "id": discussion_id
                        }
                    }
                }
            }
        }

        try:
            response_data = self.session.post(f"{self.forum_url}/api/posts", json=template).json() # type: dict

            if 'errors' in response_data:
                raise FlarumError(response_data)

            post = FlarumPost(response_data)

        except:
            raise

        finally:
            return post


    # TODO: get_all_users params + get_all_users_detailed
    def get_all_users(self) -> Union[FlarumUsers, None]:
        users = None

        try:
            url = self.ENDPOINTS['api_users_url']
            response_data = self.session.get(url=url).json() # type: dict

            if 'errors' in users:
                raise FlarumError(response_data)

            users = FlarumUsers(response_data)

        except:
            raise

        finally:
            return users


    def get_user_by_id(self, user_id: Union[str, int]) -> Union[dict, None]:
        user = None

        try:
            url = f"{self.forum_url}/api/users/{str(user_id)}"
            ud = self.session.get(url=url).json() # type: dict

            if 'errors' in ud:
                return None

            user = FlarumUser(ud)

        except:
            raise

        finally:
            return user


    def get_user_by_username(self, user_username: str) -> Union[FlarumUser, None]:
        user = None

        try:
            url = f"{self.ENDPOINTS['api_users_url']}?filter%5Bq%5D={user_username}"
            ud = self.session.get(url=url).json() # type: dict

            if 'errors' in ud:
                raise Exception(ud)

            for user in ud['data']:

                if user['attributes']['username'] == user_username:
                    user_details = self.session.get(url=f"{self.ENDPOINTS['api_users_url']}/{user['id']}").json()
                    user = FlarumUser(user_details)

                else:
                    continue

        except:
            raise

        finally:
            return user


    # TODO: Finish (filters, includes, tags, etc.) + programatic URL builder (look for URL parsers)
    def get_discussions(self, page=1, sort_by: str="lastPostedAt", include: list=None, filter: str=None) -> Union[FlarumDiscussions, None]:
        discussions = None
        parameters = {}

        if page:
            parameters['page[offset]'] = str(page * 20)

        if sort_by:
            parameters['sort'] = sort_by

        if include and type(include) == list:
            parameters['include'] = ",".join(include)
        
        if filter:
            parameters['filter'] = filter

        try:
            url = f"{self.forum_url}/api/discussions" 
            dd = self.session.get(url=url, params=parameters).json() # type: dict

            if 'errors' in dd:
                return None

            discussions = FlarumDiscussions(dd)

        except:
            raise

        finally:
            return discussions


    def get_discussion_by_id(self, discussion_id: Union[str, int]) -> Union[FlarumDiscussion, None]:
        discussion = None

        try:
            url = f"{self.forum_url}/api/discussions/{str(discussion_id)}"
            raw_discussion = self.session.get(url=url).json() # type: dict

            if 'errors' in raw_discussion:
                raise FlarumError(raw_discussion)

            discussion = FlarumDiscussion(raw=raw_discussion['data'])

        except:
            raise

        finally:
            return discussion




    def get_post_by_id(self, post_id: Union[str, int]) -> Union[FlarumPost, None]:
        post = None

        try:
            url = f"{self.forum_url}/api/posts/{str(post_id)}"
            raw_post = self.session.get(url=url).json() # type: dict

            if 'errors' in raw_post:
                raise FlarumError(raw_post)

            post = FlarumPost(raw_post['data'])

        except:
            raise

        finally:
            return post


    def vote_on_poll(self, poll: FoF_Poll, option: FoF_PollOption) -> FoF_Poll:

        url = f"{self.forum_url}/api/fof/polls/{poll.id}/vote"
        response = self.session.patch(url, json={ "data": { "optionId": str(option.id) } }).json()
        
        if 'errors' in response:
            return response
        
        poll = FoF_Poll(response['data'])
        
        for included in response['included']:

                if included['type'] == "poll_options":
                    poll['_options'] = []
                    poll['_options'].append(included)
                    continue
        
        return poll



from pyflarum.classes.extensions.Flarum_Tags import FlarumTagUserMixin


# TODO: Create 'FlarumSelfUser', inheriting this + self user properties only
class FlarumUser(FlarumTagUserMixin, dict):
    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)


    # Main:
    @property
    def id(self) -> str:
        try:
            return self['id']
        except KeyError:
            return None

    @property
    def type(self) -> str:
        try:
            return self['type']
        except KeyError:
            return None


    # Attributes:
    @property
    def attributes(self) -> dict:
        try:
            return self['attributes']
        except KeyError:
            return None

    @property
    def username(self) -> str:
        try:
            return self.attributes['username']
        except KeyError:
            return None

    @property
    def displayName(self) -> str:
        try:
            return self.attributes['displayName']
        except KeyError:
            return None

    @property
    def avatarUrl(self) -> str:
        try:
            return self.attributes['avatarUrl']
        except KeyError:
            return None

    @property
    def slug(self) -> str:
        try:
            return self.attributes['slug']
        except KeyError:
            return None



import logging
from typing import Union

import requests
from requests_cache.core import install_cache

logger = logging.getLogger(__name__)


class FlarumSession:
    def __init__(self, forum_url: str, username: Union[str, None]=None, password: Union[str, None]=None, cache: bool=True, cache_expire_after: int=300, proxies: dict=None):
        if cache:
            install_cache('./pyflarum_cache', backend='sqlite', expire_after=cache_expire_after)

        # Main:
        self.session = requests.Session()
        self.forum_url = forum_url

        # Proxies:
        if type(proxies) == dict:
            self.session.proxies.update(proxies)

        # Authentication:
        self.username = username
        self.password = password

        try:
            # Authenticating with auth data:
            self.identification_data = {"identification": self.username, "password": self.password}
            self.token_data = self.session.post(url=f'{self.forum_url}/api/token', json=self.identification_data).json() # type: dict
            
            if 'token' in self.token_data:
                self.session.headers.update(
                    {
                        "Authorization": f'Token {self.token_data["token"]}; userId={str(self.token_data["userId"])}',
                        "User-Agent": "pyflarum-bot-authenticated"
                    }
                )

                self.isAuthenticated = True
            else:
                self.isAuthenticated = False

        except Exception as e:
            # Authentication failed, fallback to "read-only" mode:
            self.identification_data = None
            self.token_data = None
            self.session.headers.update(
                {
                    "User-Agent": "pyflarum-bot-unauthenticated"
                }
            )

            self.isAuthenticated = False
            logger.error(e)
        
        if self.isAuthenticated:
            logger.info("[pyFlarum] Successfuly authenticated.")
        else:
            logger.info(f"[pyFlarum] Couldn't authenticate '{self.username}' for '{self.forum_url} - not logged in.")



from pyflarum.exceptions import FlarumError


class FlarumTagUserMixin(dict):
    def __init__(self, *args):
        """
            Initializes the data of the mixin.
        """

        dict.__init__(self, *args)


    @property
    def followTagsPageDefault(self) -> int:
        try:
            return self['data']['attributes']['preferences']['followTagsPageDefault']
        except KeyError:
            return None


class FlarumPostDiscussionTagsData(dict):
    def __init__(self, tag_ids: list=None, raw: dict=None):
        """
            Initializes the data of the mixin.
        """

        if raw is None:
            if tag_ids is None or type(tag_ids) != list:
                raise FlarumError("tag_ids parameter must be a list of tag ids.")

            tag_data = { "data": [] }
            for tag_id in tag_ids:
                tag_data['data'].append({ "type": "tags", "id": str(tag_id)})

            dict.__init__(self, tag_data)

        else:
            dict.__init__(self, raw)



from datetime import datetime

from pyflarum.date_conversions import flarum_to_datetime

class FlarumPost(dict):
    def __init__(self, *args):
        dict.__init__(self, *args)
    

    # Main:
    @property
    def id(self) -> str:
        try:
            return self['id']
        except KeyError:
            return None

    @property
    def type(self) -> str:
        try:
            return self['type']
        except KeyError:
            return None


    # Attributes:
    @property
    def number(self) -> int:
        try:
            return self['attributes']['number']
        except KeyError:
            return None

    @property
    def content_type(self) -> str:
        try:
            return self['attributes']['contentType']
        except KeyError:
            return None
    
    @property
    def content_html(self) -> str:
        try:
            if self.content_type == "comment":
                return self['attributes']['contentHtml']
        except KeyError:
            return None
    
    @property
    def content(self) -> str:
        try:
            if self.content_type == "comment":
                return self['attributes']['content']
        except KeyError:
            return None
    
    @property
    def editedAt(self) -> datetime:
        try:
            return flarum_to_datetime(self['attributes']['editedAt'])
        except KeyError:
            return None
    
    # Permissions:
    @property
    def canEdit(self) -> bool:
        try:
            return self['attributes']['canEdit']
        except KeyError:
            return None

    @property
    def canDelete(self) -> bool:
        try:
            return self['attributes']['canDelete']
        except KeyError:
            return None

    @property
    def canHide(self) -> bool:
        try:
            return self['attributes']['canHide']
        except KeyError:
            return None
    
    @property
    def isApproved(self) -> bool:
        try:
            return self['attributes']['isApproved']
        except KeyError:
            return None
    
    @property
    def canApprove(self) -> bool:
        try:
            return self['attributes']['canApprove']
        except KeyError:
            return None
    
    @property
    def canFlag(self) -> bool:
        try:
            return self['attributes']['canFlag']
        except KeyError:
            return None
    
    @property
    def canLike(self) -> bool:
        try:
            return self['attributes']['canLike']
        except KeyError:
            return None

    @property
    def canBanIP(self) -> bool:
        try:
            return self['attributes']['canBanIP']
        except KeyError:
            return None
    
    # Relationships:
    @property
    def relationships(self) -> dict:
        try:
            return self['relationships']
        except KeyError:
            return None

    @property
    def author(self) -> dict:
        try:
            return self.relationships['user']['data']
        except KeyError:
            return None

    @property
    def author_id(self) -> dict:
        try:
            return self.author['id']
        except KeyError:
            return None

    @property
    def discussion(self) -> dict:
        try:
            return self.relationships['discussion']['data']
        except KeyError:
            return None

    @property
    def discussion_id(self) -> dict:
        try:
            return self.discussion['id']
        except KeyError:
            return None
    
    @property
    def likes(self) -> dict:
        try:
            return self.relationships['likes']['data']
        except KeyError:
            return None

    @property
    def mentioned_by(self) -> dict:
        try:
            return self.relationships['mentionedBy']['data']
        except KeyError:
            return None



from pyflarum.client.classes.FlarumSession import FlarumSession
from pyflarum.client.classes.FlarumMyUser import FlarumMyUser

from pyflarum.client.classes.flarum.FlarumForum import FlarumForum

from pyflarum.client.classes.flarum.FlarumUser import FlarumUser
from pyflarum.client.classes.flarum.FlarumUsers import FlarumUsers
from pyflarum.client.classes.flarum.FlarumGroups import FlarumGroup, FlarumGroups

from pyflarum.client.classes.flarum.FlarumDiscussions import FlarumDiscussion, FlarumDiscussions
from pyflarum.client.classes.flarum.FlarumPosts import FlarumPost

from pyflarum.client.classes.extensions.Flarum_Tags import FlarumPostDiscussionTagsData, FlarumTagUserMixin
from pyflarum.client.classes.extensions.FriendsOfFlarum_Polls import FoF_Poll, FoF_PollOption


if __name__ == "__main__":
    print(
        FoF_Poll,
        FoF_PollOption,
        FlarumPostDiscussionTagsData,
        FlarumTagUserMixin,
        FlarumPost,
        FlarumDiscussion,
        FlarumDiscussions,
        FlarumGroup,
        FlarumGroups,
        FlarumUser,
        FlarumUsers,
        FlarumForum,
        FlarumMyUser,
        FlarumSession
    )



from typing import Generator


class FlarumGroup(dict):
    def __init__(self, *args):
        dict.__init__(self, *args)
    

    # Main:
    @property
    def id(self) -> str:
        try:
            return self['id']
        except KeyError:
            return None

    @property
    def type(self) -> str:
        try:
            return self['type']
        except KeyError:
            return None
    
    @property
    def nameSingular(self) -> str:
        try:
            return self['attributes']['nameSingular']
        except KeyError:
            return None

    @property
    def namePlural(self) -> str:
        try:
            return self['attributes']['namePlural']
        except KeyError:
            return None
    
    @property
    def color(self) -> str:
        try:
            return self['attributes']['color']
        except KeyError:
            return None

    @property
    def icon(self) -> str:
        try:
            return self['attributes']['icon']
        except KeyError:
            return None

    @property
    def isHidden(self) -> bool:
        try:
            return self['attributes']['isHidden']
        except KeyError:
            return None



class FlarumGroups(dict):
    def __init__(self, *args):
        dict.__init__(self, *args)


    @property
    def rawData(self) -> dict:
        try:
            return self
        except KeyError:
            return None

    @property
    def all_groups(self) -> Generator[FlarumGroup, None, None]:
        try:
            for data in self['data']:
                group = FlarumGroup(data)
                yield group
        except KeyError:
            return None



from typing import Iterable, Union
from datetime import datetime

from pyflarum.date_conversions import datetime_to_flarum, flarum_to_datetime
from pyflarum.exceptions import FlarumError


class FoF_PollOption(dict):
    """
        Represents a FoF poll option.
    """


    # Main:
    @property
    def type(self) -> Union[str, None]:
        return self.get('type', None)


    @property
    def id(self) -> Union[int, None]:
        return self.get('id')


    # Attributes:
    @property
    def attributes(self) -> dict:
        return self.get('attributes', {})


    @property
    def answer(self) -> Union[str, None]:
        return self.attributes.get('answer', None)

    @property
    def createdAt(self) -> datetime:
        try:
            return flarum_to_datetime(self['attributes']['createdAt'])
        except KeyError:
            return None

    @property
    def updatedAt(self) -> datetime:
        try:
            return flarum_to_datetime(self['attributes']['updatedAt'])
        except KeyError:
            return None


class FoF_Poll(dict):
    """
        A subclass of `dict` representing poll's JSON data.

        - `question (str)`: The poll's question (example: `"Do you like ducks?"`)
        - `options (list)`: A list containing all options for a poll (example: `["Yes", "No"]`).
        - `endDate (datetime)`: The date when the poll ends and no more votes can be cast.
        - `public (bool)`: Whether or not the poll is a public one (eg.: people can view who voted, defaults to `False`).
        # 
        - `raw_data (dict)`: The raw poll data, used when fetching the poll from an API. Either specify all the options above
        that will build the poll data for you, or just this one with already built poll data.
    """

    def __init__(self, question: str=None, options: list=None, endDate: datetime=None, public: bool=False, raw_data: dict=None):
        """
            Builds the poll data.

            - `question (str)`: The poll's question (example: `"Do you like ducks?"`)
            - `options (list)`: A list containing all options for a poll (example: `["Yes", "No"]`).
            - `endDate (datetime)`: The date when the poll ends and no more votes can be cast.
            - `public (bool)`: Whether or not the poll is a public one (eg.: people can view who voted, defaults to `False`).
            # 
            - `raw_data (dict)`: The raw poll data, used when fetching the poll from an API. Either specify all the options above
            that will build the poll data for you, or just this one with already built poll data.
        """
        
        if raw_data is None:
            if question is None or options is None or type(options) != list:
                raise FlarumError('A question and at least 1 option is required for polls. Options must also be a list')

            poll_data = {
                "question": question,
                "publicFoF_Poll": public,
                "relationships": {
                    "options": options
                }
            }

            if endDate is not None:
                poll_data['endDate'] = datetime_to_flarum(endDate)

            dict.__init__(self, poll_data)

        else:
            dict.__init__(self, raw_data)


    # Main:
    @property
    def type(self) -> str:
        try:
            return self['type']
        except KeyError:
            return None

    @property
    def id(self) -> int:
        try:
            return self['id']
        except KeyError:
            return None

    # Attributes:
    @property
    def question(self) -> str:
        try:
            return self['attributes']['question']
        except KeyError:
            return None

    @property
    def hasEnded(self) -> Union[bool, None]:
        return self.get('attributes', {}).get('hasEnded', None)

    @property
    def public(self) -> Union[bool, None]:
        self.get('attributes', {}).get('public_poll', None)

    @property
    def endDate(self) -> Union[datetime, None]:
        raw = self.get('attributes', {}).get('endDate', None)
        return flarum_to_datetime(raw)

    @property
    def createdAt(self) -> Union[datetime, None]:
        raw = self.get('attributes', {}).get('createdAt', None)
        return flarum_to_datetime(raw)

    @property
    def updatedAt(self) -> Union[datetime, None]:
        raw = self.get('attributes', {}).get('updatedAt', None)
        return flarum_to_datetime(raw)

    @property
    def relationship_options(self) -> list:
        return self.get('relationships', {}).get('options', {}).get('data', {})

    @property
    def relationship_votes(self) -> list:
        return self.get('relationships', {}).get('votes', {}).get('data', {})

    @property
    def _options(self) -> Iterable[FoF_PollOption]:
        try:
            options = []

            for option in self['_options']:
                options.append(FoF_PollOption(raw=option))

            return options

        except KeyError:
            return None

    def get_poll_option_by_id(self, id: Union[str, int]) -> FoF_PollOption:
        poll_option = None

        try:
            for option in self._options:
                if option['id'] == str(id):
                    poll_option = FoF_PollOption(option)

        except KeyError:
            raise

        finally:
            return poll_option
